##### Configuration

nolog ?= $(test)
exts ?= csv xml
test_n ?= 2
tablesSort ?= plots organisms stems

##### Vars/functions

selfDir_uZPPqC := $(dir $(lastword $(MAKEFILE_LIST)))

# Make
SHELL := /bin/bash
subMake = $(MAKE) $(@F) --directory=$(@D)
+_ = $(+:_%=)
addBeforeExt = $(basename $(2))$(1)$(suffix $(2))

# System
date = $(shell date +"%Y-%m-%d-%H-%M-%S")

# Terminal
esc := '['
reset := $(esc)'0m'
emph := $(esc)'7m '
endEmph := ' '$(reset)

# Commands
CP = cp -p
DIFF = diff --unified=2

# Paths
root := $(selfDir_uZPPqC)..
mappings := $(root)/mappings
psqlAsBien := $(root)/bin/psql_vegbien

##### General targets

all: _always maps ;

.SUFFIXES:

_always:
.PHONY: _always

clean: _always
	$(RM) $(all)

remake: _always clean all ;

%.out: %.make _always
	./$* >$@
.PRECIOUS: %.out

$(root)/%: _always
	+$(subMake)

##### SVN

svn_props: _always
	svn propset svn:ignore $$'.~*\n*.log\n*.out\nsrc*' .
	$(if $(wildcard test/),svn propset svn:ignore $$'*.out\n*.xml' test)

##### Installation

reinstall: _always uninstall install ;

##### Maps

allViaMaps := $(filter-out maps/VegBIEN.%.csv,$(wildcard maps/*.csv))
via := $(firstword $(sort $(basename $(basename $(notdir $(allViaMaps))))))

coreMap := $(mappings)/$(via)-VegBIEN.%.csv

viaMaps := $(wildcard $(tablesSort:%=maps/$(via).%.csv))
viaMaps += $(filter-out $(viaMaps),$(wildcard maps/$(via).*.csv))

autogenMaps := $(subst $(via).,VegBIEN.,$(viaMaps))
directMaps := $(autogenMaps) $(filter-out $(autogenMaps),\
$(wildcard maps/VegBIEN.*.csv))
tables := $(directMaps:maps/VegBIEN.%.csv=%)

maps: $(directMaps) _always ;

.PRECIOUS: $(allViaMaps) $(directMaps)

all += $(autogenMaps)

maps/VegBIEN.%.csv: maps/$(via).%.csv $(coreMap)
	$(root)/bin/join_sort <$+ >$@
.PRECIOUS: maps/VegBIEN.%.csv $(coreMap)

##### Mapping

dbFile := $(firstword $(wildcard src/db.*.sql))
inputFiles := $(wildcard src/*.csv src/*.xml)

+maps = $(filter maps/% $(mappings)/%,$(+_))
<in = $(firstword $(filter-out $(+maps),$(+_)) $(wildcard $(exts:%=src/*.$*.%)))
map = $(if $(<in),<$(<in) ,$(if $(mapEnv),env $(mapEnv) ))$(root)/map $(+maps)
map2db = env out_database=vegbien $(map)

##### Import to VegBIEN

ifneq ($(dbFile)$(inputFiles),)

import: _always import-all verify ;

log = $(@:-all=)$(if $(n),.n=$(n),).$(date).log
import = -(set -x; "time" env commit=1 $(map2db)) $(if $(nolog),,\
$(if $(n),,>>$(log))) 2>&1$(if $(nolog),,$(if $(n),|tee -a $(log)))
# ignore import errors, which are often benign (e.g. invalid date format)

ifneq ($(inputFiles),)
import-all: $(addprefix import-,$(tables)) _always ;
else
import-all: $(directMaps) _always
	$(import)
endif

import-%: maps/VegBIEN.%.csv _always
	$(import)
# default:
import-%: _always ;

else
import: _always ;
endif

##### Log files from import

logs := $(wildcard *.log)

rm_logs: _always
	$(RM) $(logs)

##### Verification of import

ifneq ($(wildcard verify.ref*),)

verify: verify.ref verify.out _always
	$(DIFF) $(+_)

all += verify.out

%.out: %.sql _always
	$(out_cmd)
.PRECIOUS: %.out
out_cmd = $(psqlAsBien) --no-align --field-separator='	' --pset=footer=off\
--pset=null=NULL <$< >$@

else
verify: _always ;
endif

##### Testing

ifneq ($(wildcard test/),)

hasOwnRef = $(filter-out %.2-step.xml,$@)
testRef = $(1:.2-step.xml=.xml).ref

define test
@echo "Testing $(abspath $@)..."
>$@ env test=1 verbose=1 n=$(test_n) $(1)
@(set -x; $(DIFF) $(call testRef,$@) $@) 2>&1 || { e=$$?;\
$(if $(wildcard $(call testRef,$@)),,cat $@;)\
$(if $(hasOwnRef),\
echo $(emph)"To accept new test output:"$(endEmph);\
echo "$(MAKE) $@-ok --directory=$(realpath .) --makefile=../input.Makefile";\
,\
echo $(emph)"Note: The preceding failed test is compared to another test's\
output"$(endEmph);\
echo $(emph)"When it fails, this always indicates a bug"$(endEmph);\
)\
exit $$e;}
endef

test2File = $(call test,$(map))

tests :=

test/$(via).%.xml: maps/$(via).%.csv _always
	$(test2File)
tests += test/$(via).%.xml

test/VegBIEN.%.xml: maps/VegBIEN.%.csv _always
	$(test2File)
tests += test/VegBIEN.%.xml

test/VegBIEN.%.2-step.xml: test/$(via).%.xml $(coreMap) _always
	-$(test2File)
# Don't abort tester if only 2-step test fails, as it's often finicky
tests += test/VegBIEN.%.2-step.xml

test/import.%.out: maps/VegBIEN.%.csv _always
	$(call test,$(map2db))
tests += test/import.%.out

.PRECIOUS: $(tests)

testOutputs := $(foreach test,$(tests),$(tables:%=$(test)))

test: _always $(testOutputs) ;

all += $(testOutputs)

# Accepts a test output: make <test_output_path>-ok
%-ok: _always
	$(CP) $* $(call testRef,$*)

else
test: _always ;
endif

##### Input-type-specific

# Each input type needs var $(mapEnv) and targets install, uninstall

ifneq ($(dbFile),)

dbEngineExt := $(subst .,,$(suffix $(basename $(notdir $(dbFile)))))
db := $(notdir $(realpath .))

%.ref: %.ref.sql
	$(inputDbAsBien) $(db) <$< >$@
.PRECIOUS: %.ref

#### Installation

install: _always db ;

uninstall: _always rm_db ;

#### DB-engine-specific

# Each DB engine needs vars $(dbEngine), $(dbAsBien) and targets db, rm_db

ifeq ($(dbEngineExt),my)

dbEngine := MySQL

bienPassword := $(shell cat $(root)/config/bien_password)
mysqlAs = mysql --user=$(1) --password='$(bienPassword)'
mysqlAsRoot := $(call mysqlAs,root)
dbAsBien := $(call mysqlAs,bien)

db: $(dbFile) _always
	echo "CREATE DATABASE $(db) DEFAULT CHARACTER SET latin1;"|$(mysqlAsRoot)
	-$(mysqlAsRoot) --database=$(db) <$<
	echo "GRANT SELECT ON $(db).* TO 'bien'@'localhost';"|$(mysqlAsRoot)
# ignore errors in db import so that GRANT will still be run

rm_db: _always
	-echo "REVOKE ALL ON $(db).* FROM 'bien'@'localhost';"|$(mysqlAsRoot)
	echo "DROP DATABASE IF EXISTS $(db);"|$(mysqlAsRoot)
# ignore errors if grant not defined

### Unrecognized DB engine

else
$(error The DB filename $(dbFile) must have the form db.{my|pg}.sql)
endif

# Must come after dbEngine is set
mapEnv := in_engine=$(dbEngine) in_database=$(db)

#### Unrecognized input type

else
install: _always ;
uninstall: _always ;
endif
